#
# BIND configuration module
#
# Jason Vas Dias, Red Hat Inc., September 2004

import os
import commands
import ftplib
import datetime
import copy
import base64
import Lookup
import RR
import DNS
from   Zone      import *
from   NamedConf import *
from   ZoneTree  import *
from   View      import *
from   Server    import *
import Options
import DNSsec

class BIND:
    isViewed=None
    
    def __init__(self, gui, argv):
        self.gui = gui
        if(  (os.access('/usr/sbin/named-checkconf',os.F_OK)!=1)
           or(os.access('/usr/sbin/named-checkzone',os.F_OK)!=1)
           or(os.access('/usr/sbin/named',os.F_OK)!=1)
           or(os.access('/usr/sbin/dns-keygen',os.F_OK)!=1)
           or(os.access('/usr/sbin/dnssec-keygen',os.F_OK)!=1)
           or(os.access('/usr/sbin/dnssec-signzone',os.F_OK)!=1)                          
          ):
            gui.error(_('Initialization Error'),_('This application requires the Red Hat BIND distribution\n(bind, bind-utils, bind-libs) RPMs to be installed.'),True
                     )
            return
        if(  (os.access('/usr/sbin/named-checkconf',os.X_OK)!=1)
           or(os.access('/usr/sbin/named-checkzone',os.X_OK)!=1)
           or(os.access('/usr/sbin/named',os.X_OK)!=1)
           or(os.access('/usr/sbin/dns-keygen',os.X_OK)!=1)
           or(os.access('/usr/sbin/dnssec-keygen',os.X_OK)!=1)
           or(os.access('/usr/sbin/dnssec-signzone',os.X_OK)!=1)             
          ):
            gui.error(_('Initialization Error:'),_('You do not have permission to run the BIND utilities\nrequired by this application:\n/usr/sbin/{named-checkconf,named-checkzone,dns-keygen}.'),
                      True
                     )
            return
        chroot=''
        r=commands.getstatusoutput('. /etc/sysconfig/named; echo "$ROOTDIR"')
        if (r[0] == 0):
            chroot=r[1]        
        if os.access(chroot + '/etc/named.conf',os.F_OK) != 1:
            if os.access(SCB + '/profiles/default/named.conf',os.F_OK)==1:
                debug( 'INIT CONFIG')
                self.gui.warning(self.gui.gui,_('Initializing to Defaults:'),_('No BIND configuration was found.\nInstalling the initial default configuration.'))
                debug( 'INITIALIZING')
                r=commands.getstatusoutput('set -e; /bin/mkdir -p '+chroot+'/{etc,var/{named/{slave,data},tmp,run/named},dev}; if [ ! -e '+chroot+'/dev/random ]; then /bin/mknod '+chroot+'/dev/random c 1 8; fi; if [ ! -e '+chroot+'/dev/null ]; then /bin/mknod '+chroot+'/dev/null c 1 3; fi; if [ ! -e '+chroot+'/dev/zero ]; then /bin/mknod '+chroot+'/dev/zero c 1 5; fi; /bin/cp -fp $SCB/profiles/default/named.conf '+chroot+'/etc; /bin/cp -fp $SCB/profiles/default/named/* '+chroot+'/var/named; /bin/chown -R root:named '+chroot+'/var/named; /bin/chown named:named '+chroot+'/var/{named/{slave,data},run/named};  /bin/chown root:root '+chroot+'/etc/named.conf; echo -ne "key rndckey {\n\talgorithm hmac-md5;\n\tsecret \""`/usr/sbin/dns-keygen`"\";\n};\n">'+chroot+'/etc/rndc.key; /bin/chown root:named '+chroot+'/etc/rndc.key; /bin/chmod 640 '+chroot+'/etc/rndc.key; /bin/chmod 777 '+chroot+'/var/tmp; if [ -e /selinux/enforce ] && [ -x /sbin/restorecon ]; then /sbin/restorecon -R /var/{named,run,tmp}; fi')
                debug(r)
                if r[0] != 0 :
                   self.gui.error(_('Initialization Error:'),_('Configuration could not be initialized from defaults:\n')+r[1],True)
                   return
                r = self.get_root_hints(filename='named.root', dir=chroot+'/var/named')
                if r != None:
                    self.gui.error(_('Initialization Error:'),_('Root Cache Download Failed:\n\t%s\nEnsure this machine is connected to the Internet.') % r,True)
            else:
                self.gui.error(_('Initialization Error'),_('No current BIND configuration was found,\nand no default configuration was found\nin %s/profiles/default.\nInstall the caching-nameserver package and restart.') % SCB,True)                        
        r=commands.getstatusoutput('/usr/sbin/named-checkconf -v')
        if (r[0] != 0) or (len(r[1]) <= 0):
            gui.error(_('Initialization Error'),
                      _('named-checkconf -v failed:\n')+str(r[0])+' :'+r[1],
                      True
                     )
            return
        self.version = r[1]
        debug( 'BIND VERSION:',self.version)


        r=commands.getstatusoutput('/bin/rpm -q bind')
        if (r[0] != 0) or (len(r[1]) <= 0):
            gui.error(_('Initialization Error'),
                      'rpm -q bind:\n'+str(r[0])+' :'+r[1],
                      True
                     )
            return
	debug( 'BIND PACKAGE:',r[1])
	self.version_main = string.split(r[1],'-')[1];

        self.nmdc = NamedConf()        
        if len(self.nmdc.errors):
            gui.error(_('DNS Configuration File %s Initialisation Error:') % self.nmdc.files[0]['file'] , string.join(self.nmdc.errors,"\n"),True)
            return
        self.zoneTree = None
        self.nmdc.bind=self
        self.load()
        
    def load(self):
        self.zones={}
        self.views={}
        self.acls={}
        self.keys={}
        self.isViewed = None
        self.acls={}
        self.servers={}
        DNS.ACL.namedACLs={}
        DNS.Key.namedKeys={}
        if self.nmdc.__contains__('server'):
            for svr in self.nmdc['server'].keys():
                self.servers[svr] = Server(self.nmdc,['server',svr])
        for origin, real_key, filename, clas, cfg, view in self.nmdc.zones():
            if self.isViewed == None:
                self.isViewed = (view != None)
            if (self.isViewed and (view == None)) or ( (not self.isViewed) and (view != None) ):
                self.gui.error(_('DNS Configuration Initialisation Error:'), _('If any views are defined, ALL zones must be in a view.'),True)
                return
            parent=['zone']
            if self.isViewed :
                parent = [ 'view', view, 'zone' ]                
            zone = Zone(filename, origin, self.nmdc, parent + [ real_key ], clas, cfg['type'])
            if len(zone.errors) != 0 :
                if (origin == '.') and (cfg['type'] == 'hint') and (os.access(filename,os.F_OK) != 1):
                    r = self.get_root_hints( dir=self.nmdc.directory())
                    if r != None:
                        self.gui.error(_('Initialization Error'),_('Root Cache Download Failed:\n\t%s\nEnsure this machine is connected to the Internet.') % r,True)
                    zone = Zone(filename, origin, self.nmdc, parent + [ real_key ], clas, cfg['type'])
                    if len(zone.errors) != 0:                        
                        self.gui.error(_('Zone File Initialisation Error:'), _('Zone %s\nfile %s\nload error:\n%s') % (origin, filename, string.join(zone.errors,"\n")), fatal=True) 
                else:
                    fatalE = False
                    ew=_('Warning')
                    for e in zone.errors:
                        if not (str(e).find('WARNING')>-1):
                            fatalE=True
                            ew=_('Error')
                            break
                    debug('FATAL:',fatalE)
                    self.gui.error(_('Zone File Initialisation %s:') % ew,_('Zone %s\nfile %s\nload error:\n%s') % (origin, filename, string.join(zone.errors,"\n")), fatal=fatalE, warning=not fatalE)
            if self.zone(origin,parent) != None:
                # both "." (hint) and "." (master zone) specified - named_checkconf
                # does not complain, named starts, but named gets really screwed
                # all other duplicate zones are not tolerated by named-checkconf & named
                self.gui.error( _('Zone File Initialisation Error:'),
                           _('ERROR: configuration contains duplicate zones:\n\tzone: %s class: %s type: %s file: %s\n\tzone: %s class: %s type: %s file: %s') %
                            ( origin, clas, cfg['type'], cfg['file'], origin,
                              self.zones[origin]['clas'], self.zones[origin]['config']['type'],
                              self.zones[origin]['config']['file'] ),
                           True
                         )
                return
            if view == None:
                self.zones[origin]={'key':real_key, 'parent':['zone'], 'zone': zone, 'clas':clas, 'config':cfg, 'file': filename}
            else:
                vk=view.strip('"')
                if not self.views.has_key(vk):
                    self.views[vk]={'zones':{}, 'keys':{}, 'servers':{}, 'key':view}
                    self.views[vk]['view'] = View(self.nmdc,view,self.views[vk])
                self.views[vk]['zones'][origin] = {'key':real_key, 'parent':['view',view,'zone'], 'zone': zone, 'clas':clas, 'config':cfg, 'file': filename, 'view':vk}
                if self.zones.has_key(origin):
                    if type( self.zones[origin] ) != list:
                        self.zones[origin] = [ self.zones[origin] ]
                    self.zones[origin].append( self.views[vk]['zones'][origin] )
                else:
                    self.zones[origin] =  self.views[vk]['zones'][origin]
        if self.isViewed:
            for vk in self.nmdc.view.keys():
                view=vk.strip('"')
                if not self.views.has_key(view):
                    self.views[view] = {'zones':{}, 'keys':{}, 'servers':{}, 'key':vk}
                    self.views[view]['view'] = View(self.nmdc,vk,self.views[view])
                if self.nmdc[['view',vk]].has_key('key'):
                    for k in self.nmdc[['view',vk,'key']].keys():
                        self.views[view]['keys'][k.strip('"')]=DNS.Key(self.nmdc[['view',vk,'key',k]],self.nmdc,['view',vk,'key',k])
                if self.nmdc[['view',vk]].has_key('server'):
                    for svr in self.nmdc[['view',vk,'server']].keys():
                        self.views[view]['servers'][svr] = Server(self.nmdc,['view',vk,'server',svr])
        if self.nmdc[''].has_key('acl'):
            for acl in self.nmdc.acl.keys():
                try:
                    self.acls[acl] = DNS.ACL( self.nmdc.acl[acl], self.nmdc, ['acl',acl] )
                except DNS.ACL.Undefined, e:
                    self.gui.error(_('DNS Configuration Initialisation Error:'),
                                   _('Access control list\n%s\ncontains a reference to an undefined ACL\n%s.') %
                                   (acl, e.name),True
                                  )
        if self.nmdc[''].has_key('key'):
            for key in self.nmdc.key.keys():
                k = key.strip('"')
                self.keys[k] = DNS.Key( self.nmdc.key[key], self.nmdc, key )
        if self.nmdc.__contains__('trusted-keys'):
            self.trusted_keys=DNSsec.TrustedKeys(self)
        if self.zoneTree == None:
            self.zoneTree = ZoneTree(self)
        else:
            self.zoneTree.load()

    def dirty(s):
        dirty=(hasattr( s.nmdc,'dirty' ) and (s.nmdc.dirty == True))
        if not dirty:
            for z in s.zones.keys():
                if( type(s.zones[z]) == list):
                    zones = s.zones[z]
                else:
                    zones = [ s.zones[z] ]
                for zone in zones:
                    dirty = dirty or (zone.has_key('dirty') and (zone['dirty'] == True))        
        return dirty

    def save(s):
        namedRunning = s.named_is_running()
        for z in s.zones.keys():
            if type(s.zones[z]) == list:
                for zn in s.zones[z]:
                    if zn.has_key('dirty') and (zn['dirty'] == True):
                        needReSign=False
                        debug( 'SAVE ZONE:', zn['zone'].zone_origin )
                        if hasattr(zn['zone'], 'additions'):
                            debug(  '   ADDITIONS:', zn['zone'].additions )
                        if zn.has_key('dnssecSecure'):
                            debug(  '    DNSSEC', s.zones[z]['dnssecSecure'] )
                        if ( hasattr(zn['zone'], 'additions') and
                             (zn['zone'].additions > 0) and
                             zn.has_key('dnssecSecure')
                           ):
                            needReSign = True
                            debug('NEED RESIGN!')
                        if namedRunning and zn['config'].has_key('allow-update'):
                            if zn.has_key('view') :
                                debug("FREEZE:", zn['zone'].zone_origin + ' ' + zn['clas'] +' '+ zn['view'])
                                commands.getstatusoutput(
                                    '/usr/sbin/rndc freeze ' + zn['zone'].zone_origin + ' ' + zn['clas'] +' '+ zn['view']
                                    )
                            else:
                                debug("FREEZE:", zn['zone'].zone_origin)
                                commands.getstatusoutput('/usr/sbin/rndc freeze ' + zn['zone'].zone_origin )
                        zn['zone'].save()
                        if namedRunning and zn['config'].has_key('allow-update'):
                            if zn.has_key('view') :
                                debug("UNFREEZE:", zn['zone'].zone_origin + ' ' + zn['clas'] +' '+ zn['view'])
                                commands.getstatusoutput(
                                    '/usr/sbin/rndc unfreeze ' + zn['zone'].zone_origin + ' ' + zn['clas'] +' '+ zn['view']
                                    )
                            else:
                                debug("UNFREEZE:", zn['zone'].zone_origin)
                                commands.getstatusoutput('/usr/sbin/rndc unfreeze ' + zn['zone'].zone_origin )
                        if len(zn['zone'].errors ) > 0:
                             s.gui.modalError(s.gui.gui, _('DNS Zone File Save Error:'),  string.join(zn['zone'].errors,"\n"))
                             return False;
                        zn['dirty'] = False
                        if needReSign:
                            debug(  'RE-SIGNING!' )
                            DNSsec.SignZone( s, s.zones[z]['zone'] )
            elif(s.zones[z].has_key('dirty') and (s.zones[z]['dirty'] == True)):
                needReSign=False
                debug(  'SAVE ZONE:', s.zones[z]['zone'].zone_origin )
                if hasattr(s.zones[z]['zone'], 'additions'):
                    debug(  '   ADDITIONS:', s.zones[z]['zone'].additions )
                if s.zones[z].has_key('dnssecSecure'):
                    debug(  '    DNSSEC', s.zones[z]['dnssecSecure'] )
                if ( hasattr(s.zones[z]['zone'], 'additions') and
                     (s.zones[z]['zone'].additions > 0) and
                     s.zones[z].has_key('dnssecSecure') and
                     (s.zones[z]['dnssecSecure'] == False)
                   ):
                    needReSign = True
                    debug(  'NEED RESIGN!' )
                zn=s.zones[z]
                if namedRunning and zn['config'].has_key('allow-update'):
                    if zn.has_key('view') :
                        debug("FREEZE:", zn['zone'].zone_origin + ' ' + zn['clas'] +' '+ zn['view'])
                        commands.getstatusoutput(
                            '/usr/sbin/rndc freeze ' + zn['zone'].zone_origin + ' ' + zn['clas'] +' '+ zn['view']
                            )
                    else:
                        debug("FREEZE:", zn['zone'].zone_origin)
                        commands.getstatusoutput('/usr/sbin/rndc freeze ' + zn['zone'].zone_origin )                    
                s.zones[z]['zone'].save()
                if namedRunning and zn['config'].has_key('allow-update'):
                    if zn.has_key('view') :
                        debug("UNFREEZE:", zn['zone'].zone_origin + ' ' + zn['clas'] +' '+ zn['view'])
                        commands.getstatusoutput(
                            '/usr/sbin/rndc unfreeze ' + zn['zone'].zone_origin + ' ' + zn['clas'] +' '+ zn['view']
                            )
                    else:
                        debug("UNFREEZE:", zn['zone'].zone_origin)
                        commands.getstatusoutput('/usr/sbin/rndc unfreeze ' + zn['zone'].zone_origin )
                if len(s.zones[z]['zone'].errors):
                    s.gui.modalError(s.gui.gui, _('DNS Zone File Save Error:'),  string.join(s.zones[z]['zone'].errors,"\n"))
                    return False
                s.zones[z]['dirty'] = False
                if needReSign:
                    debug(  'RE-SIGNING!' )
                    DNSsec.SignZone( s, s.zones[z]['zone'] )
        s.nmdc.save()
        if len(s.nmdc.errors):
            s.gui.modalError(s.gui.gui, _('DNS Configuration File Save Error:'), string.join(s.nmdc.errors,"\n"))
            return False
        return True
    
    def zone(s, origin, parent):
        if (not s.zones.has_key(origin)) and len(origin) and (origin != '.') and (origin[-1] == '.'):
            origin = origin[:-1]
        if not s.isViewed:
            if s.zones.has_key(origin):
                return s.zones[origin]
        else:
            if s.zones.has_key(origin):
                if type(s.zones[origin]) == list:
                    for z in s.zones[origin]:
                        if( z['parent'] == parent):
                            return z
                else:
                    if( s.zones[origin]['parent'] == parent):
                        return s.zones[origin]
        return None
    
    def get_zone_node(s, zone):
        origin = zone.zone_origin
        if (not s.zones.has_key(origin)) and len(origin) and (origin[-1] == '.'):
            origin = origin[:-1]
        if not( s.zones.has_key(origin)):
            return None
        if type(s.zones[origin]) == list:
            for z in s.zones[origin]:
                if z['zone'] == zone:
                    return z
        else:
            return s.zones[origin]

    class NullZone:
        origin=''
        zone_origin=''
        type=None
        clas=None
        records=[]
        ttl=None
        def In(s,arg):
            return True    
        zTypes=[[_('Forward'), DNS.FORWARD, DNS.Name],
            [_('IPV4 Reverse'), DNS.IPV4.REVERSE, DNS.IPV4.ReverseAddress],
            [_('IPV6 Reverse'), DNS.IPV6.REVERSE_ARPA, DNS.IPV6.ReverseARPAddress],
            [_('NSAP Reverse'), DNS.NSAP.REVERSE, DNS.NSAP.ReverseAddress]
           ]
        type=DNS.FORWARD

    def server_name(s):
        me = Lookup.localhost.name
        if me[-1] != '.':
            me += '.'
        return me
    
    def soa(s, zone, name):
        host = s.server_name()
        return SOA( zone, name,None,None,
                       [ '@',
                         'SOA',
                         host,
                         'root.' + host,
                         1,
                         '3h', '1h', '1w', '1h'
                       ], [],2
                     )

    def ns(s, zone, name, soa):
        return NS(zone, name, None, None, [ '@', 'NS', soa.MNAME.value() ], [],2)
        
    def zone_file(s, zone, name, unique=None, zoneType='master'):
        fname=name
        if( fname == '.'):
            fname = 'root'
        if zone.type == DNS.IPV4.REVERSE:
            fname=DNS.IPV4.ReverseAddress(fname,zone).label()
        elif zone.type == DNS.IPV6.REVERSE_ARPA:
            fname=DNS.IPV6.ReverseARPAddress(fname,zone).label()
        elif zone.type == DNS.IPV6.REVERSE_INT:
            fname=DNS.IPV6.ReverseINTAddress(fname,zone).label()
        elif zone.type == DNS.NSAP.REVERSE:
            fname=DNS.NSAP.ReverseAddress(fname,zone).label()
        if zoneType in [ 'slave', 'stub' ]:
            fname = 'slaves/' + fname;            
        debug( 'FNAME: ZONE TYPE:', zone.type, 'FNAME:',fname)
        if( fname[-1] == '.'):
            fname = fname[:-1]
        if unique != None:            
            f=s.nmdc.directory()+'/'+fname+'.db'
            n=0
            newfn=''
            while os.access(f,os.F_OK)==1:
               newfn=fname+'_'+str(n)
               f=s.nmdc.directory()+'/'+newfn+'.db'
               n+=1
            if(newfn != ''):
                fname=newfn
        return fname+'.db'
    
    def create_zone(s, nmdcKey, clas, origin, zoneType, cfgType, filename=None, soa=None, ns=None, options=None):
        origin_key='"' + origin + '"' + ' ' +clas
        if s.nmdc.__contains__(nmdcKey + ['zone']):                    
            nmdcKey += ['zone',  origin_key  ]
            parent = nmdcKey[:-1]
        else:
            nmdcKey += ['zone']
            parent = nmdcKey
        debug( 'CREATE ZONE - PARENT:', parent, 'TYPE:',cfgType, 'ORIGIN:',origin,'FILE:',filename)
        fname=None
        if( cfgType != 'forward' ) and ( cfgType != 'delegation-only' ):
            if(filename == None):
                zn=BIND.NullZone()
                zn.type = zoneType
                fname = s.zone_file(zn,origin,True)
            else:
                fname = filename
        dir = s.nmdc.directory()
        if (fname != None) and  (len(fname) > len(dir)) and (fname[:len(dir)] == dir):
            fname=fname[:len(dir)+1]
        zfName=None
        if (fname != None):
            if fname[0] != '/':
                zfName = dir + '/' + fname
        zone = Zone(zfName, origin, s.nmdc, nmdcKey, clas, cfgType)
        if cfgType not in [ 'forward', 'slave', 'stub', 'delegation-only' ]:
            zone.ttl = soa.MINIMUM
            s.nmdc[ nmdcKey ] =  'zone '+ origin_key +' ' + '{\n'  + '\ttype ' + cfgType + ';\n' + '\tfile "' + fname + '";\n' + "};\n"
        elif cfgType == 'forward':
            if not options.has_key('forwarders'):
                return
            s.nmdc[ nmdcKey ] =  'zone '+ origin_key +' ' + '{\n'  + '\ttype ' + cfgType + ';\n\t' + 'forwarders '+options['forwarders'].out() + '};\n'
        elif cfgType == 'delegation-only':
            s.nmdc[ nmdcKey ] =  'zone '+ origin_key +' ' + '{\n'  + '\ttype ' + cfgType + ';\n' + "};\n"
        elif cfgType in [ 'slave', 'stub' ]:
            s.nmdc[ nmdcKey ] =  'zone '+ origin_key +' ' + '{\n'  + '\ttype ' + cfgType + ';\n' + '\tfile ' + '"' + fname + '";\n\t' + 'masters ' + options['masters'].out() + '};\n'
        s.nmdc.update()
        zone.records=[]
        if( soa != None ):
            zone.records.append(soa)
            zone.SOA=soa
            soa.zone = zone
        if( ns != None):
            zone.records.append(ns)
            ns.zone = zone
        if nmdcKey[0] == 'view':
            vk=nmdcKey[1].strip('"')
            debug( 'PARENT + origin_key:', parent+[origin_key])
            s.views[vk]['zones'][zone.origin] = {'key':origin_key, 'parent':parent, 'zone':zone, 'clas':clas, 'config':s.nmdc[parent+[origin_key]], 'file': filename, 'dirty':True, 'view':vk }
            e = s.zone(zone.origin,parent)
            if (e != None):
                if (type(e) != list):
                    s.zones[ zone.origin ] = [e]
                s.zones[ zone.origin ].append(s.views[vk]['zones'][zone.origin])
            else:
                s.zones[ zone.origin ] = s.views[vk]['zones'][zone.origin]
        else:
            s.zones[zone.origin] = {'key':origin_key, 'parent':['zone'], 'zone':zone, 'clas':clas, 'config':s.nmdc[nmdcKey], 'file': filename, 'dirty':True}
        s.zoneTree.load()
    
    def add_host(s,ip,dn,cn=None,view=None,forward_only=False,reverse_only=False):
        debug( 'ADD HOST:',ip,dn,reverse_only)
        parent=[]
        if view != None:
            zones = view.v['zones']
            parent= ['view',view.key]
        else:
            zones = s.zones
        zoneParent=parent[:]
        zoneForward=None
        zoneReverse=None
        nullZone = BIND.NullZone()
        maxDnMatch=0
        maxIpMatch=0
        if( dn[-1] != '.'):
            dn += '.'
        name = DNS.Name(dn,nullZone)
        ipv4=None
        ipv6=None
        nsap=None
        if (len(ip)>=2) and (ip[0:2] == '0x'):
            nsap = DNS.NSAP.Address(ip)
            if (nsap.address==None) or (len(nsap.address) <= 0):
                debug( 'Unrecognized Address')
                return _('Unrecognized Address')
        if nsap == None:
            ipv4=DNS.IPV4.Address(ip,input=True)
            ipv6=None
            rtyp=None
            if(ipv4.errors != None):
                ipv4 = None            
                ipv6 = DNS.IPV6.Address(ip)
                if ipv6.address == None:
                    debug( 'Unrecognized Address')
                    return (-1,_('Unrecognized address: ')+ip)
                reverse = DNS.IPV6.reverse( ipv6.nibble_out() ) + '.' + DNS.IPV6.REVERSE_ARPA_ZONE
                addr = DNS.IPV6.ReverseARPAddress(reverse,nullZone)
                rtype= DNS.IPV6.REVERSE_ARPA
            else:
                reverse=DNS.IPV4.reverse(ip) + '.' + DNS.IPV4.REVERSE_ZONE
                addr = DNS.IPV4.ReverseAddress(reverse,nullZone)
                rtype= DNS.IPV4.REVERSE
        else:
            reverse = DNS.NSAP.reverse( ip ) + '.' + DNS.NSAP.REVERSE_ZONE
            addr = DNS.NSAP.ReverseAddress(reverse,nullZone)
            rtype= DNS.NSAP.REVERSE
        for origin in zones.keys():
            if( zones[origin]['zone'].cfgType != 'master'):
                debug( '   NOT MASTER ZONE:',origin)
                continue
            zone_origin = zones[origin]['zone'].zone_origin
            debug( '    COMPARE ZONE:',zone_origin,dn,zones[origin]['zone'].type            )
            if(    (zones[origin]['zone'].type == DNS.FORWARD)
               and (( zone_origin != '.') or (len(DNS.split_dots(dn)) == 1))
               and ( len(dn) >= len(zone_origin))
               and ( dn[-len(zone_origin):] == zone_origin )              
               and ( len(zone_origin) > maxDnMatch )
              ):
                maxDnMatch = len(zone_origin)
                zoneForward=origin
                debug( '    MATCH FORWARD')
            elif( (   ( ipv4 != None )
                   and( zones[origin]['zone'].type == DNS.IPV4.REVERSE)                    
                   and( zones[origin]['zone'].SOA.Name().contains(ip))
                   and( len(zone_origin) > maxIpMatch )
                  )
                or(   ( ipv6 != None)
                   and( (zones[origin]['zone'].type == DNS.IPV6.REVERSE_ARPA)
                      or(zones[origin]['zone'].type == DNS.IPV6.REVERSE_INT)
                      )
                   and( zones[origin]['zone'].SOA.Name().contains(str(addr)) )
                   and( len(zone_origin) > maxIpMatch )
                  )
                or(   ( nsap != None)
                   and( zones[origin]['zone'].type == DNS.NSAP.REVERSE)
                   and( zones[origin]['zone'].SOA.Name().contains(ip))
                   and( len(zone_origin) > maxIpMatch )
                  )
                ):
                maxIpMatch = len(zone_origin)
                zoneReverse=origin
                debug( '    MATCH REVERSE')
        forwardZN=None
        reverseZN=None
        for dnsName, ztype, z in( (name, DNS.FORWARD, zoneForward), (addr, rtype, zoneReverse) ):
            parent=zoneParent[:]
            if( z == None ):
                if (ztype == DNS.FORWARD) and reverse_only:
                    continue
                elif (ztype != DNS.FORWARD) and forward_only:
                    continue                
                nullZone.type = ztype
                (dnName, dnOrig)=dnsName.split_domain()
                if( dnOrig == '.' ):
                    # FAKE TLD zone (eg. 'localhost.')
                    if dnName[-1] != '.':
                        dnName=dnName + '.'
                    dnOrig=dnName
                nullZone.origin = dnOrig                    
                debug( 'ADD HOST:', dnName, dnOrig, nullZone.type                    )
                nullZone.zone_origin =  nullZone.origin
                orig=dnsName.__class__('@',nullZone)
 #               print 'NAME:', orig.value()
                nullZone.records=[]                
                soa = s.soa(nullZone, copy.deepcopy(orig))
 #               print 'SOA NAME:',soa.name
                ns = s.ns(nullZone, copy.deepcopy(orig), soa)
 #               print 'NS NAME:',ns.name
                fname = s.zone_file(nullZone,dnOrig,unique=True)
                debug( 'CREATE ZONE:', dnOrig, fname)
                s.create_zone(parent, 'IN', dnOrig, ztype, 'master', fname, soa, ns)
                z=zones[dnOrig]['zone']
                if( ztype == DNS.FORWARD ):
                    zoneForward =  z
                    forwardZN=zones[dnOrig]
                else:
                    zoneReverse =  z
                    reverseZN=zones[dnOrig]
                soa.zone=z
                ns.zone=z
                z.records=[]
                z.records.append(soa)
                z.records.append(ns)
            else:
                if( ztype == DNS.FORWARD ):
                    forwardZN=zones[zoneForward]
                    zoneForward = zones[zoneForward]['zone']
                else:
                    reverseZN=zones[zoneReverse]
                    zoneReverse = zones[zoneReverse]['zone']
        name.origin = zoneForward.origin
        name.zone_origin = zoneForward.zone_origin
        addr.origin = zoneReverse.origin
        addr.zone_origin = zoneReverse.zone_origin
#        print 'HOST:', name.relative_label()
#        print 'ADDR:', addr.value()
#        print 'FORWARD SOA ORIG:', zoneForward.SOA.Name().zone_origin,zoneForward.SOA.Name().value() 
#        print 'REVERSE SOA ORIG:', zoneReverse.SOA.Name().zone_origin,zoneReverse.SOA.Name().value()
        if not reverse_only:
            foundRR=None
            if ipv6 == None:
                for r in zoneForward.records:
                    if (r.Type()=='A') and (str(r.Name())==str(name)):                        
                        foundRR=r
                        break
                if foundRR==None:                    
                    A(zoneForward, name, None, None, [ name.relative_label(), 'A', ip ],[])
                    forwardZN['dirty']=True
                else:
                    if( addr.label() != foundRR.ADDRESS.label() ):
                        if( foundRR.ADDRESS.addr != ip):
                            foundRR.ADDRESS.addr = ip
                            forwardZN['dirty']=True
            else:
                for r in zoneForward.records:
                    if (r.Type()=='AAAA') and (str(r.Name())==str(name)):                        
                        foundRR=r
                        break
                if foundRR==None:
                    AAAA(zoneForward, name, None, None, [ name.relative_label(), 'AAAA', addr.address.hex_out() ],[])
                    forwardZN['dirty']=True
                else:
                    if( addr.label() != foundRR.ADDRESS.label() ):
                        foundRR.ADDRESS = addr
                        forwardZN['dirty']=True               
        if not forward_only:
            a=str(zoneReverse.SOA.Name())
            b=str(addr)
            addr=addr.__class__(b[:-(len(a)+1)],zoneReverse)
            foundRR=None
            for r in zoneReverse.records:
                if (r.Type()=='PTR') and (str(r.Name())==str(addr)):                        
                    foundRR=r
                    break
            if( foundRR == None ):                
                PTR(zoneReverse, addr, None, None, [ addr.value(), 'PTR', str(name) ],[])
                reverseZN['dirty']=True
            else:
                if( str(foundRR.PTRDNAME) != str(name)):
                    foundRR.PTRDNAME = name
                    reverseZN['dirty']=True
        s.zoneTree.load()

    def copyZone(s, zone, fromView, toView):
        debug(  'copyZone:',zone,fromView,toView )
        if not ( (fromView != toView) and
                 s.views.has_key(fromView) and s.views.has_key(toView)
                 and s.views[fromView]['zones'].has_key(zone) and
                 ( not(s.views[toView].has_key('zones')) or not(s.views[toView]['zones'].has_key(zone)))
               ):
            debug(  'NO!', 'have from view:',s.views.has_key(fromView), 'have to view:',  s.views.has_key(toView),                  'fromView has zone:', s.views[fromView]['zones'].has_key(zone), 'toView has zones:',                   s.views[toView].has_key('zones'), 'toView has zone:', s.views[toView]['zones'].has_key(zone) )
            return
        fromView = s.views[fromView]
        toView=    s.views[toView]
        zone = fromView['zones'][zone]
        oldKey = zone['zone'].nmdcKey
        newKey = []
        if s.nmdc.__contains__(['view',toView['key'],'zone']):
            newKey = ['view',toView['key'],'zone',zone['key']]
        else:
            newKey = ['view',toView['key'],'zone']
        zoneText = s.nmdc.out(oldKey)
        debug(  'OLD ZONE TEXT:', zoneText )
        s.nmdc[ newKey ] = zoneText
        s.nmdc.update()
        f = s.nmdc[ ['view', fromView['key'], 'zone', zone['key'], 'file'] ]
        f = f.strip('"')
        old_file=s.nmdc.directory() + '/' + f
        f =  toView['key'].strip('"')  + '_' + f
        if os.access( s.nmdc.directory() + '/' + f, os.F_OK )==1:
            try:
                os.rename(s.nmdc.directory() + '/' + f, s.nmdc.directory() + '/' + f + '.' + str(time.time()))
            except OSError,e:
                debug( 'ERROR on rename', s.nmdc.directory() + '/' + f,' to ',s.nmdc.directory() + '/' + f + '.' + str(time.time()),':', e)
        try:
            out_fd=os.open(s.nmdc.directory() + '/' + f,os.O_WRONLY | os.O_CREAT, 0644)
            in_fd=os.open(old_file,os.O_RDONLY)
            r=os.read(in_fd,512)
            while len(r):
                os.write(out_fd,r)
                r=os.read(in_fd,512)
            os.fsync(out_fd)
            os.close(out_fd)
            os.close(in_fd)
            commands.getstatusoutput( "if [ -x /sbin/restorecon ]; then /sbin/restorecon "+s.nmdc.directory() + '/' + f + "; fi" )
        except OSError, e:
            debug( 'ERROR copy ', old_file, ' to ',s.nmdc.directory() + '/' + f)
        s.nmdc[ [ 'view', toView['key'], 'zone', zone['key'], 'file'] ]= '"' + f + '"'            
        s.load()

    def moveZoneFile(s, zone, newLocation):
        failed=False
        if os.access( zone.filename, os.F_OK ) == 1:
            try:
                os.rename( zone.filename,  s.nmdc.directory() + '/' + newLocation )
            except OSError,e:
                failed=True
                debug( 'ERROR on rename', zone.filename,' to ',s.nmdc.directory() + '/' + newLocation +' : ', e)
        if not failed:
            zone.filename = newLocation
            s.nmdc [ zone.nmdcKey ]['file']='"'+newLocation+'"'
            if newLocation[:7] == 'slaves/':
                commands.getstatusoutput('chown named:named '+newLocation+'; if [ -x /sbin/restorecon ]; then /sbin/restorecon '+newLocation+'; fi')
            return True
        return False

    def acl_users(s, aclName):
        users=[]
        for n in s.acls.keys():
            if s.acls[n].contains_acl(aclName):
                users.append(str(['acl',+s.acls[n].name]))
        options = Options.Options(s.nmdc)
        for o in options.keys():
            ov=options[o]
            if ov.value.__class__ == DNS.ACL:
                if ov.value.contains_acl(aclName):
                    users.append(str(['options',o]))
        if s.isViewed:
            for v in s.views.keys():
                if s.views[v]['view'].inputFilter.contains_acl(aclName):
                    users.append(str(s.views[v]['view'].nmdcKey + ['match-clients']))
                if s.views[v]['view'].outputFilter.contains_acl(aclName):
                    users.append(str(s.views[v]['view'].nmdcKey + ['match-destinations']))
                options = Options.Options(nmdc=s.nmdc,obj=s.views[v]['view'])
                for o in options.keys():
                    ov=options[o]
                    if ov.value.__class__ == DNS.ACL:
                        if ov.value.contains_acl(aclName):
                            users.append(str(s.views[v]['view'].nmdcKey + [o]))
                for zk in s.views[v]['zones'].keys():
                    zone=s.views[v]['zones'][zk]['zone']
                    options = Options.Options(nmdc=s.nmdc,obj=zone)
                    for o in options.keys():
                        ov=options[o]
                        if ov.value.__class__ == DNS.ACL:
                            if ov.value.contains_acl(aclName):
                                users.append(str(s.views[v]['view'].nmdcKey + ['zone',zk, o]))
        else:
            for z in s.zones.keys():
                zone=s.zones[z]['zone']
                options = Options.Options(nmdc=s.nmdc,obj=zone)
                for o in options.keys():
                    ov=options[o]
                    if ov.value.__class__ == DNS.ACL:
                        if ov.value.contains_acl(aclName):
                            users.append(str(['zone',zk, o]))
        return users
    
    def zone_origins(s,ztype=None,view=None):       
        if view != None:
            view=view.strip('"')
            zones = s.views[view]['zones']
        else:
            zones=s.zones
        origins=[]
        for z in zones.keys():
            zone=None
            if type(zones[z]) == list:
                if (view==None):
                    zone=zones[z][0]['zone']
                else:
                    for zn in zones[z]:
                        if (view==zn['view'] ) or (view.strip('"')==zn['view']):
                            zone=zn['zone']
                            break
            else:
                zone=zones[z]['zone']
            if (zone != None) and ((ztype==None) or (ztype==zone.type)):
                origins.append(zone)
        return origins

    def query_root_serial(s):
        s,a=commands.getstatusoutput('dig +noall +nocmd +nocomments +answer +tries=2 +time=1 @a.root-servers.net. . soa')
        debug("query root serial:",s,a)
        if ( s == 0 ) and len(a):
            zone=Zone(None, '.', NullNmdc(), ['zone','.'])
            zone.contents=a
            zone.parse()
            if hasattr(zone,'SOA'):
                return str(zone.SOA.SERIAL)
            else:
                debug("query root serial: parse failed:",zone.errors,zone.records)
        return None
            
    def query_root_hints(s,filename=None,dir=None):
        cwd=None
        if( dir != None):
            try:
                cwd=os.getcwd()
                os.chdir(dir)
            except OSError:
                dir=None
                cwd=None                
        if(filename==None):
            filename='named.root'
        serial=s.query_root_serial()
        if( serial != None ):
            s,a=commands.getstatusoutput('dig +noall +nocmd +nocomments +answer +additional +tries=2 +time=1 @a.root-servers.net. . ns')
            if ( s == 0 ):
                d=datetime.datetime.now()
                ts=str(d.year)+'-'+str(d.month)+'-'+str(d.day)+'_'+str(d.hour)+'.'+str(d.minute)+'.'+str(d.second)
                try:
                    if os.access(filename,os.F_OK):
                        os.rename(filename,filename+'.'+ts)
                except OSError:
                    pass                    
                try:
                    fd=os.open(filename,os.O_WRONLY | os.O_CREAT,0644)
                    os.write(fd,
                              '; root "." zone hints file, queried of a.root-servers.net. by system-config-bind\n'
                             +'; version of root zone: '+ serial +'\n'
                             + a
                             +'\n'
                             )
                    os.fsync(fd)
                    os.close(fd)
                    commands.getstatusoutput('if [ -x /sbin/restorecon ]; then /sbin/restorecon '+os.getcwd()+'/'+filename+'; fi')
                except OSError,e:
                    try:
                        if os.access(filename+'.'+ts,os.F_OK):                
                            os.rename(filename+'.'+ts,filename)
                    except OSError:
                        pass
                    if(cwd != None):
                        try:
                            os.chdir(cwd)
                        except OSError:
                            pass
                    return str(e[1])
            else:
                if(cwd != None):
                    try:
                        os.chdir(cwd)
                    except OSError:
                        pass
                return str(s)+a
        else:
            if(cwd != None):
                try:
                    os.chdir(cwd)
                except OSError:
                    pass
            return "Query of root SOA failed."
        if(cwd != None):
            try:
                os.chdir(cwd)
            except OSError:
                pass
        return None
    
    def ftp_root_hints(s,filename=None,dir=None):
        cwd=None
        if( dir != None):
            try:
                cwd=os.getcwd()
                os.chdir(dir)
            except OSError:
                dir=None
                cwd=None                
        if(filename==None):
            filename='named.root'
        d=datetime.datetime.now()
        ts=str(d.year)+'-'+str(d.month)+'-'+str(d.day)+'_'+str(d.hour)+'.'+str(d.minute)+'.'+str(d.second)
        try:
            if os.access(filename,os.F_OK):
                os.rename(filename,filename+'.'+ts)
        except OSError:
            pass
        try:
            # prevent 5 minute timeout when DNS servers aren't responding ...
            nic = Lookup.Lookup('ftp.rs.internic.net')
            if nic.addr == None:
                nic.addr = '198.41.0.6' # last known address - unlikely to change!
            debug( 'ROOT HINTS DOWNLOAD:','"'+nic.addr+'"')
            ftp = ftplib.FTP(nic.addr)
        except ftplib.all_errors, e:
            try:
                if os.access(filename+'.'+ts,os.F_OK):                
                    os.rename(filename+'.'+ts,filename)
            except OSError:
                pass
            debug( 'ftplib error:', e[1] )
            if(cwd != None):
                try:
                    os.chdir(cwd)
                except OSError:
                    pass
            return str( e[1] )
        try:
            ftp.login()
            ftp.retrbinary('RETR domain/named.root',open(filename,'wb').write)
        except ftplib.all_errors,e:
            try:
                if os.access(filename+'.'+ts,os.F_OK):                
                    os.rename(filename+'.'+ts,filename)
            except OSError:
                pass
            debug( 'ftp error:', e[1] )
            if(cwd != None):
                try:
                    os.chdir(cwd)
                except OSError:
                    pass
            return str( e[1] )
        commands.getstatusoutput('chmod 644 '+filename+'; if [ -x /sbin/restorecon ]; then /sbin/restorecon '+filename+'; fi')
        if(cwd != None):
            try:
                os.chdir(cwd)
            except OSError:
                pass
        return None

    def get_root_hints(s, filename=None,dir=None):
        e = s.query_root_hints(filename, dir)
        if e != None:
            debug("Root Zone query failed:",e)
            e = s.ftp_root_hints(filename,dir)
            if e != None:
                debug("Root Zone FTP failed:",e)
        return e
    
    def newKey(s):
        r = commands.getstatusoutput('/usr/sbin/dns-keygen')
        if r[0] != 0:
            return [ -1, r[1] ]
        return r[1]

    def version_cmp(s,ver):
        vn=ver.split('.')
        vc=s.version.split('.')
        for v in vc :
            if len(vn) == 0:
                return -1
            vt=vn.pop(0)
            try:
                vic = int(v)
                vin = int(vt)
                if vic > vin:
                    return -1
                elif vic < vin:
                    return 1
            except ValueError:
                return 0
        return 0

    def start_named(s):
        if os.access('/var/lock/subsys/named',os.F_OK)==1:
            return (-1,'named is already running')
        return commands.getstatusoutput('/sbin/service named start')

    def stop_named(s):
        if os.access('/var/lock/subsys/named',os.F_OK)==0:
            return (-1,'named is not running')
        return commands.getstatusoutput('/sbin/service named stop')

    def restart_named(s):
        return commands.getstatusoutput('/sbin/service named condrestart')

    def reload_named(s):
        if os.access('/var/lock/subsys/named',os.F_OK)==0:
            return commands.getstatusoutput('/sbin/service named start')
        else:
            return commands.getstatusoutput('/sbin/service named reload')

    def named_is_running(s):
        return (os.access('/var/lock/subsys/named',os.F_OK)==1)
