#!/usr/bin/python
#
# BIND named.conf Options
#
# Jason Vas Dias, Sept. 2004

import DNS
from   NamedConf        import NamedConf
from   NamedConfOptions import O
from   OptionCheck      import OptionCheck
from   View             import View

class Option:
    value={}
    option={}
    def __init__(s, opt, val):
        s.option = opt
        s.value=val

class Options:
    
    def parse(s):
        s.opts={}
        s.objKeys=[]
        try:
            s.obj = s.nmdc[ s.key ]
            if( type(s.key) != list):
                s.objKeys = [s.key]
            else:
                s.objKeys = s.key[:]
        except IndexError:
            got_obj=False
            if (len(s.type)>0):
                try:
                    if( type(s.key) != list ):
                        nmdcKey = [ s.type, s.key ]
                    else:
                        nmdcKey = [ s.type ] + s.key
                    s.obj = s.nmdc[ nmdcKey ]
                    s.objKeys = nmdcKey
                    got_obj=True
                except IndexError:
                    debug( 'No nmdc key:', nmdcKey)
                    debug( s.nmdc[''])
                    return
            if not got_obj:
                debug( 'No nmdc obj: k:', s.key)
                debug( s.nmdc[''])
                return
        if type(s.obj) != dict :
            debug( 'Unhandled nmdc obj type:', type(s.obj))
            return
        for k in s.obj.keys():
            debug( 'OPT:',k)
            keys = k.split()
            key =  keys[0]
            val = s.obj[k]
            if not O.has_key(key):
                continue
            else:
                if O[key].has_key('type'):
                    if (O[key]['type'] == DNS.ACL) or (O[key]['type'] == DNS.SortList):
                        debug( 'OPTION ACL:',s.objKeys + [key])
                        s.opts[key]={ 'obj': s.obj[k], 'key':k, 'opt':key, 'option': O[key], 'value':O[key]['type'](s.obj[k],s.nmdc, s.objKeys + [key] ) }
                    elif(O[key]['type'] == DNS.Radio):
                        s.opts[key]={ 'obj': s.obj[k], 'key':k, 'opt':key, 'option': O[key], 'value':DNS.Radio(s.obj[k],O[key]['values']) }
                    else:
                        s.opts[key]={ 'obj': s.obj[k], 'key':k, 'opt':key, 'option': O[key], 'value':O[key]['type'](s.obj[k]) }
                elif O[key].has_key('attrs'):
                    debug( 'ATTRS OPTION ',key)
                    vals=[]
                    value={}
                    if len(keys) > 1:
                        vals = keys[1:]
                    if type(val) == list:
                        vals += val
                    else:
                        vals.append(val)
                    synKeys = []
                    synKeys = []
                    for attrd in O[key]['attrs']:
                        for name, attr in attrd.items():
                            if attr.has_key('syntax'):
                                synKeys.append(name)
                    for attrd in O[key]['attrs']:
                        for name, attr in attrd.items():
                            if value.has_key(name):
                                continue
                            if( len(vals)):
                                debug( '   ATTR', name, vals[0])
                                if (type(vals[0])==str) and (vals[0].lower() in synKeys) and (len(vals)>1):
                                    aname = vals.pop(0)
                                    v=vals.pop(0)
                                    if( aname != name ):
                                        debug( '    ATTR SKIP',name)
                                        value[name]=None
                                        for a in O[key]['attrs']:
                                            for n,at in a.items():
                                                if n==aname:
                                                    attr=at
                                                    break
                                        name=aname
                                    debug( '    ATTR <',name,'> =',v)
                                elif( attr.has_key('syntax') ):
                                    debug( '    ATTR SKIP',name)
                                    value[name]=None
                                    continue
                                else:
                                    v=vals.pop(0)
                                if( attr['type'] == DNS.ACL):
                                    value[name]=attr['type'](v,s.nmdc, s.objKeys + [key])
                                elif(attr['type'] == DNS.Radio):
                                    value[name]=attr['type'](v,attr['values'])
                                elif(attr['type'] == DNS.IPV4.Address):
                                    value[name]=DNS.IPV4.Address(v,input=True,prefixAllowed=(v.find('/')>0))
                                elif(attr['type'] == DNS.IPV6.Address):
                                    value[name]=DNS.IPV6.Address(v,allow_prefix=(v.find('/')>0))
                                else:
                                    value[name]=attr['type'](v)
                                debug( '    ATTR:', name, '=', value[name])
                            else:
                                break
                            debug( '   ATTR NEXT')
                    s.opts[key]={ 'obj': s.obj[k], 'key':k, 'opt':key, 'option': O[key], 'value':value }
    
    def __init__(s, nmdc=None, key=None, obj=None):
        s.opts={}
        s.types=['zone','view','server']
        s.changes={}    
        if(  type(nmdc) == type(None) ):
            nmdc = NamedConf()
            if( len(nmdc.errors) != 0):
                return
        s.nmdc = nmdc
        s.obj = obj
        s.type=''
        s.key = None
        if (type(obj) != type(None))  and (hasattr(obj,'__class__')):
            s.type = str(obj.__class__).split('.')[-1].lower()
            if s.type == 'zonenode':
                s.type = 'zone'
                s.key = obj['zone'].nmdcKey
                s.zone = obj['zone']
                s.zoneType = obj['zone'].cfgType
            elif s.type == 'zone':
                s.type = 'zone'
                s.key = obj.nmdcKey
                s.zone = obj
                s.zoneType = obj.cfgType
        if s.key == None:
            if( key == None ):
                s.key = 'options'
            else:
                s.key = key
        debug( 'TYPE:',s.type)
        debug( 'KEY:',s.key)
        s.parse()
        debug( 'KEY AFTER PARSE:',s.key)
        s.checkOptions = OptionCheck(O)
        
    def all_options(s):        
        if( len(s.type) == 0):
            a=O.keys()
            a.sort()
            for ok in a:
                global_option=True
                o = O[ok]
                for t in s.types:
                    if o.has_key(t+'-only'):
                        global_option=False
                        break
                if global_option and (not o.has_key('deprecated')) and not s.opts.has_key(ok) and (
                    not o.has_key('unimplemented')) and (not o.has_key('todo')) and (
                    (not o.has_key('version')) or (hasattr(s.nmdc,'bind') and (
                     s.nmdc.bind.version_cmp(o['version'])<=0) ) ):
                    yield ok, o
        else:
            zoneType = None
            if s.type == 'zone':
                zoneType = s.zoneType
            a=O.keys()
            a.sort()
            for ok in a:
                o = O[ok]
                if (o.has_key(s.type) or o.has_key(s.type +'-only') or (
                    (s.type == 'view') and not ( o.has_key('zone-only') or o.has_key('server-only')))) and (
                    not s.opts.has_key(ok)) and (not o.has_key('deprecated')) and (not o.has_key('unimplemented')) and (not o.has_key('todo')) and (
                    (zoneType == None) or (zoneType in o['zone-types'])) and(
                    (not o.has_key('version')) or (hasattr(s.nmdc,'bind') and (s.nmdc.bind.version_cmp(o['version'])<=0)
                    )):
                    yield ok, o

    def add(s, k):
        if (not O.has_key(k)) or s.opts.has_key(k):
            debug( 'ADD OPTION FAIL:',k,s.opts.has_key(k))
            return [-1,_('Internal Program Error')]
        r = O[k]['add'](s)
        if r != OptionCheck.OK:
            debug( 'ADD OPTION CHECK FAILED:',k)
            return r
        if not O[k].has_key('type'):
            if( O[k].has_key('attrs') ):
                opt={}
                for attrd in O[k]['attrs']:
                    for name, attr in attrd.items():
                        if attr.has_key('default'):
                            val=attr['default']
                        else:
                            val=None
                        if (attr['type'] == DNS.ACL) or (attr['type'] == DNS.SortList):
                            opt[name]=attr['type']({ val : None }, s.nmdc)
                        elif(attr['type'] == DNS.Radio):
                            opt[name]=DNS.Radio(val,attr['values'])
                        elif(attr['type'] == int):
                            if( val == None ):
                                opt[name] = 0
                            else:
                                opt[name] = int(val)
                        elif(attr['type'] == str):
                            if( val == None ):
                                opt[name] = ''
                            else:
                                opt[name] = str(val)
                        elif(attr['type'] == DNS.IPV4.Address):
                            opt[name]=DNS.IPV4.Address(val,input=True,prefixAllowed=True)
                        elif(attr['type'] == DNS.IPV6.Address):
                            opt[name]=DNS.IPV6.Address(val,allow_prefix=True)
                        else:
                            opt[name]=attr['type'](val)
                s.opts[k] = { 'opt':k, 'option': O[k], 'key':k, 'value': opt }
                debug( 'ADD ATTR OPTION',k,'=',opt)
            else:
                debug( 'OPTION ',k,'ERROR: NO TYPE OR ATTRS')
                return [-1,_('Internal Program Error')]
        else:
            if O[k].has_key('default'):
                val=O[k]['default']
            else:
                val=None
            if (O[k]['type'] == DNS.ACL) or (O[k]['type'] == DNS.SortList):
                s.opts[k]={  'opt':k, 'option': O[k], 'key':k, 'value':O[k]['type'](val, s.nmdc) }
            elif(O[k]['type'] == DNS.Radio):
                s.opts[k]={ 'opt':k, 'option': O[k], 'key':k, 'value':DNS.Radio(val,O[k]['values']) }
            else:
                if( O[k]['type'] == DNS.TTL ):
                    debug( 'TTL OPTION:',k, val)
                s.opts[k]={ 'opt':k, 'option': O[k], 'key':k, 'value':O[k]['type'](val) }
        debug( 'ADDED OPTION ',k,'OK:',s.opts[k])
        s.changes[k]={'type':'ADD','opt': s.opts[k]}
        return None
    
    def remove(s, k):
        if not O.has_key(k) or not s.opts.has_key(k):
            return [-1,_('Internal Program Error')]
        r = O[k]['remove'](s)
        if r != OptionCheck.OK:
            return r
        s.changes[k]={'type':'REMOVE', 'opt':s.opts[k]}
        del s.opts[k]
        debug( 'REMOVED OPTION: ',k, s.opts.has_key(k) )
        return None

    def change(s, k):
        if not O.has_key(k) or not s.opts.has_key(k):
            return [-1,_('Internal Program Error')]
        if( hasattr(s.opts[k]['value'],'value')):
            r = O[k]['valid'](s, s.opts[k]['value'].value())
        else:
            r = O[k]['valid'](s, s.opts[k]['value'])
        if r != OptionCheck.OK:
            return r
        if not(s.changes.has_key(k) and (s.changes[k]['type'] == 'ADD')) :            
            s.changes[k]={'type':'CHANGE', 'opt':s.opts[k]}
        return None

    def attr_option_value(s,opt):
        v=opt['value']
        o=opt['option']
        s=''
        for attrd in o['attrs']:
            for name, attr in attrd.items():
                if v.has_key(name):
                    if attr.has_key('syntax'):
                        s += name + ' '
                    s+=str(v[name]) + ' '        
        return s[:-1]+';'

    def update(s):
        debug( 'UPDATE OPTIONS', s.key)
        if ( s.key == None ):
            nmdcKey = [ 'options' ]
        elif type(s.key) != list:
            nmdcKey = [ s.key ]
        else:
            nmdcKey=s.key
        if not s.nmdc.__contains__('options'):
            s.nmdc[ 'options' ] = 'options {\n};\n';
            s.nmdc.update();
        changed=False
        debug( 'UPDATE KEY:',nmdcKey)
        for k in s.changes.keys():
            c = s.changes[k]
            o = c['opt']
            if c['type'] == 'ADD':
                if type(o['value']) == dict:
                    v = s.attr_option_value(o)
                else:
                    if( hasattr(o['value'],'out')):                        
                        v = o['value'].out()
                    else:
                        v = str(o['value'])+';'
                debug( '  NEW', s.type, o['key'],v)
                s.nmdc[ nmdcKey + [o['key']] ] = o['key'] + ' ' + v + "\n"
                changed = True
            elif c['type'] == 'REMOVE':
                debug( '  DEL', s.type, o['key'])
                del s.nmdc[ nmdcKey + [o['key']] ]
                changed = True
            elif c['type'] == 'CHANGE':
                if( type(o['value']) == dict ):
                    v=s.attr_option_value(o)
                else:
                    if( hasattr(o['value'],'out')):                        
                        v = o['value'].out()
                    else:
                        v = str(o['value'])+';'
                debug( '  SET', s.type, o['key'],v)
                if type( s.nmdc[ nmdcKey + [o['key']] ] ) == dict:
                    debug( 'SET - DEL FIRST')
                    del s.nmdc[ nmdcKey + [o['key']] ]                    
                    s.nmdc[ nmdcKey + [o['key']] ] = o['key'] +' '+ v
                else:
                    i=v.rfind(';')
                    if i > -1:
                        v=v[:i] + '\n'
                    debug( 'SET')
                    s.nmdc[ nmdcKey + [o['key']] ] = v           
                changed = True
        if changed:
            s.nmdc.update()
            s.changes={}
                
    def __hasattr__(s, attr):
        return s.opts.has_key(attr)
    
    def __getattr__(s, attr):
        if attr == 'opts':
            return s.opts
        if s.opts.has_key(attr):
            return Option(s.opts[attr]['option'],s.opts[attr]['value'])
        return None
        
    def __delattr__(s, attr):
        if s.opts.has_key(attr):
            del s.opts[attr]
        
    def __getitem__(s, key):
        return s.__getattr__(key)

    def __setitem__(s, key, val):
        if s.opts.has_key(key):
            s.opts[key]['value']=s.opts[key]['type'](val)

    def __delitem__(s, key):
        s.__delattr__(key)

    def __contains__(s,key):
        return s.__hasattr__(key)

    def keys(s):
        a=s.opts.keys()
        a.sort()
        for k in a:
            yield k
    
    def __str__(s):
        return s.key

if __name__=='__main__':
    opts=Options(None)
    for k in opts.keys():
        debug( k, opts[k])
